unit ProcessorAffinity;

interface

uses
  Classes, SysUtils, Windows, SyncObjs;

type
  TAffinityMaskEnum = 0..31;
  TAffinityMaskSet = set of TAffinityMaskEnum;
  TOpenThreadDLLProc = function (dwDesiredAccess: LongWord; InheritHandle: Boolean; ThreadId: LongWord): THandle; stdcall;
  TProcessorAffinity = class(TComponent)
  private
  	FProcessAffinity: Boolean;
    FThreadAffinity: Boolean;
    FExceptionOnError: Boolean;
    FLastErrorMessage: string;
    FSetProcessMaskOnLoad: Boolean;
    FSystemMask: TAffinityMaskSet;
    FProcessMask: TAffinityMaskSet;
    PHandle: THandle;
    PID: LongWord;
    FThreadMasks: TStringList;
    CriticalSection: TCriticalSection;
    OpenThread: TOpenThreadDLLProc;
    DLLHandle: THandle;
    procedure SetThreadAffinityMask(const ID: LongWord; const Mask: LongWord; const ExceptionOnError: Boolean);
    function NewProcessThreadList: TList;
    function GetThreadIDs(I: Integer): LongWord;
    function GetThreadMasks(I: Integer): TAffinityMaskSet;
    procedure SetThreadMasks(I: Integer; Value: TAffinityMaskSet);
    function GetThreadMasksFromID(ID: LongWord): TAffinityMaskSet;
    procedure SetThreadMasksFromID(ID: LongWord; Value: TAffinityMaskSet);
    procedure SetBooleanReadOnly(const Value: Boolean);
    procedure SetSystemMaskReadOnly(const Value: TAffinityMaskSet);
    procedure SetProcessMask(Value: TAffinityMaskSet);
    function GetThreadCount: Integer;
  protected
    procedure Loaded; override;
  public
    property ThreadCount: Integer read GetThreadCount;
    property ThreadIDs[I: Integer]: LongWord read GetThreadIDs;
    property ThreadMasks[I: Integer]: TAffinityMaskSet read GetThreadMasks write SetThreadMasks;
    property ThreadMasksFromID[ID: LongWord]: TAffinityMaskSet read GetThreadMasksFromID write SetThreadMasksFromID;
    procedure AddThread(const ID: LongWord; Mask: TAffinityMaskSet = []);
    procedure RemoveThread(const ID: LongWord);
    procedure AddAllProcessThreads;
    procedure RemoveAllThreads;
    procedure SpreadThreads;
    procedure Lock;
    procedure Unlock;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
  	property ProcessAffinity: Boolean read FProcessAffinity write SetBooleanReadOnly;
    property ThreadAffinity: Boolean read FThreadAffinity write SetBooleanReadOnly;
    property ExceptionOnError: Boolean read FExceptionOnError write FExceptionOnError;
    property LastErrorMessage: string read FLastErrorMessage;
    property SetProcessMaskOnLoad: Boolean read FSetProcessMaskOnLoad write FSetProcessMaskOnLoad;
    property SystemMask: TAffinityMaskSet read FSystemMask write SetSystemMaskReadOnly;
    property ProcessMask: TAffinityMaskSet read FProcessMask write SetProcessMask;
  end;
  EProcessorAffinity = class(Exception);
  EThreadNotFound = class(EProcessorAffinity);
  EDuplicateThread = class(EProcessorAffinity);

procedure Register;

implementation

uses
  shellApi, tlhelp32;

const
  THREAD_ALL_ACCESS = $0001F03FF;
  THREAD_SET_INFORMATION = $00000020;

resourcestring
  EThreadNotFoundMsg = 'Thread not found';
  EDuplicateThreadMsg = 'Thread already present';

//function OpenThread; external 'kernel32.dll' name 'OpenThread';
//function OpenThread(dwDesiredAccess: LongWord; InheritHandle: Boolean; ThreadId: LongWord): THandle; stdcall;

procedure Register;
begin
  RegisterComponents('Not Visible', [TProcessorAffinity]);
end;

procedure TProcessorAffinity.SetThreadAffinityMask(const ID: LongWord; const Mask: LongWord; const ExceptionOnError: Boolean);
var
  Handle: THandle;
begin
  Handle := OpenThread(STANDARD_RIGHTS_REQUIRED or THREAD_ALL_ACCESS,false,ID);
  try
    if Windows.SetThreadAffinityMask(Handle,Mask) <> 0 then
      FLastErrorMessage := ''
    else begin
      FLastErrorMessage := SysErrorMessage(GetLastError);
      if ExceptionOnError then
        raise EProcessorAffinity.CreateFmt('%s - %s: %s',[Self.ClassName,'SetThreadAffinityMask',FLastErrorMessage]);
    end;
  finally
    CloseHandle(Handle);
  end;
end;

function TProcessorAffinity.NewProcessThreadList: TList;
var
 Snap: THandle;
 TE32: TTHREADENTRY32;
begin
  Result := TList.Create;;
  Snap := 0;
  try
    Snap := CreateToolhelp32Snapshot(TH32CS_SNAPTHREAD,0);
    if Snap <> 0 then begin
      TE32.dwSize := SizeOf(TTHREADENTRY32);
      if Thread32First(Snap,TE32) then begin
        if TE32.th32OwnerProcessID = PID then
          Result.Add(Pointer(TE32.th32ThreadID));
        while Thread32Next(Snap,TE32) do
          if TE32.th32OwnerProcessID = PID then
            Result.Add(Pointer(TE32.th32ThreadID));
      end;
    end;
  finally
    CloseHandle(Snap);
  end;
end;

procedure TProcessorAffinity.SetBooleanReadOnly(const Value: Boolean);
begin
  ;
end;

function TProcessorAffinity.GetThreadCount: Integer;
begin
  FLastErrorMessage := '';
  CriticalSection.Enter;
  Result := FThreadMasks.Count;
  CriticalSection.Leave;
end;

function TProcessorAffinity.GetThreadIDs(I: Integer): LongWord;
begin
  FLastErrorMessage := '';
  CriticalSection.Enter;
  try
    Result := LongWord(StrToInt(FThreadMasks.Strings[I]));
  finally
    CriticalSection.Leave;
  end;
end;

function TProcessorAffinity.GetThreadMasks(I: Integer): TAffinityMaskSet;
var
  Mask: LongWord;
begin
  FLastErrorMessage := '';
  Result := [];
  if FThreadAffinity then begin
	  CriticalSection.Enter;
    try
      Mask := LongWord(FThreadMasks.Objects[I]);
    finally
      CriticalSection.Leave;
    end;
	  Move(Mask,Result,SizeOf(Mask));
	end
	else
		Result := FProcessMask;
end;

procedure TProcessorAffinity.SetThreadMasks(I: Integer; Value: TAffinityMaskSet);
var
  Mask: LongWord;
  ID: LongWord;
begin
  FLastErrorMessage := '';
  ID := 0;
  if FThreadAffinity then begin
	  CriticalSection.Enter;
	  Value := Value * FProcessMask;
	  if Value = [] then
	    Value := FProcessMask;
	  Move(Value,Mask,SizeOf(Mask));
    try
      FThreadMasks.Objects[I] := TObject(Mask);
      ID := StrToInt(FThreadMasks.Strings[I]);
    except
      CriticalSection.Leave;
    end;
    if ID > 0 then
      Self.SetThreadAffinityMask(ID,Mask,FExceptionOnError);
  end;
end;

function TProcessorAffinity.GetThreadMasksFromID(ID: LongWord): TAffinityMaskSet;
var
  I: Integer;
  Mask: LongWord;
begin
  FLastErrorMessage := '';
  Result := [];
  if FThreadAffinity then begin
	  CriticalSection.Enter;
	  try
	    if FThreadMasks.Find('$' + IntToHex(ID,8),I) then
	      Mask := LongWord(FThreadMasks.Objects[I])
	    else begin
		    FLastErrorMessage := EThreadNotFoundMsg;
    		if FExceptionOnError then
      		raise EThreadNotFound.CreateFmt('%s - %s',[Self.ClassName,FLastErrorMessage]);
	    end;
	  finally
	    CriticalSection.Leave;
	  end;
	  Move(Mask,Result,SizeOf(Mask));
	end
	else
		Result := FProcessMask;
end;

procedure TProcessorAffinity.SetThreadMasksFromID(ID: LongWord; Value: TAffinityMaskSet);
var
  I: Integer;
  Mask: LongWord;
begin
  FLastErrorMessage := '';
  if FThreadAffinity then begin
	  CriticalSection.Enter;
	  Value := Value * FProcessMask;
	  if Value = [] then
	    Value := FProcessMask;
	  Move(Value,Mask,SizeOf(Mask));
	  try
	    if FThreadMasks.Find('$' + IntToHex(ID,8),I) then
	      FThreadMasks.Objects[I] := TObject(Mask)
	    else begin
		    FLastErrorMessage := EThreadNotFoundMsg;
    		if FExceptionOnError then
      		raise EThreadNotFound.CreateFmt('%s - %s',[Self.ClassName,FLastErrorMessage]);
	    end;
	  finally
	    CriticalSection.Leave;
	  end;
	  if FLastErrorMessage = '' then
	    Self.SetThreadAffinityMask(ID,Mask,FExceptionOnError);
  end;
end;

procedure TProcessorAffinity.SetSystemMaskReadOnly(const Value: TAffinityMaskSet);
begin
  ;
end;

procedure TProcessorAffinity.SetProcessMask(Value: TAffinityMaskSet);
var
  I: TAffinityMaskEnum;
begin
  FLastErrorMessage := '';
  if FProcessAffinity then begin
	  Value := Value * FSystemMask;
	  I := Low(TAffinityMaskEnum);
	  while Value = [] do
	    if I in FSystemMask then
	      Include(Value,I);
    CriticalSection.Enter;
 	  FProcessMask := Value;
    CriticalSection.Leave;
	  if not (csDesigning in ComponentState) and not (csLoading	in ComponentState) then begin
	    if not SetProcessAffinityMask(PHandle,LongWord(Value)) then begin
	      FLastErrorMessage := SysErrorMessage(GetLastError);
	      if ExceptionOnError then
	        raise EProcessorAffinity.CreateFmt('%s - %s: %s',[Self.ClassName,'SetProcessAffinityMask',FLastErrorMessage]);
	    end;
	  end;
	end;
end;

procedure TProcessorAffinity.AddThread(const ID: LongWord; Mask: TAffinityMaskSet);
var
  I: Integer;
begin
  CriticalSection.Enter;
  Mask := Mask * FProcessMask;
  if Mask = [] then
    Mask := FProcessMask;
  try
    if not FThreadMasks.Find('$' + IntToHex(ID,8),I) then begin
      FThreadMasks.AddObject('$' + IntToHex(LongWord(ID),8),TObject(LongWord(Mask)));
		  if FThreadAffinity then
		    Self.SetThreadAffinityMask(ID,LongWord(Mask),FExceptionOnError);
    end
    else begin
      FLastErrorMessage := EDuplicateThreadMsg;
      if FExceptionOnError then
        raise EDuplicateThread.CreateFmt('%s - %s',[Self.ClassName,FLastErrorMessage]);
    end;
  finally
    CriticalSection.Leave;
  end;
end;

procedure TProcessorAffinity.RemoveThread(const ID: LongWord);
var
  I: Integer;
begin
  CriticalSection.Enter;
  try
    if FThreadMasks.Find('$' + IntToHex(ID,8),I) then begin
			FThreadMasks.Delete(I);
		  if FThreadAffinity then
		    Self.SetThreadAffinityMask(ID,LongWord(FProcessMask),FExceptionOnError);
    end
    else begin
	    FLastErrorMessage := EThreadNotFoundMsg;
  		if FExceptionOnError then
    		raise EThreadNotFound.CreateFmt('%s - %s',[Self.ClassName,FLastErrorMessage]);
    end;
  finally
    CriticalSection.Leave;
  end;
end;

procedure TProcessorAffinity.AddAllProcessThreads;
var
  IDs: TList;
  I: Integer;
begin
  FLastErrorMessage := '';
  IDs := NewProcessThreadList;
  CriticalSection.Enter;
  try
    if FThreadAffinity then
	    for I := 0 to Pred(FThreadMasks.Count) do
  	    Self.SetThreadAffinityMask(LongWord(StrToInt(FThreadMasks.Strings[I])),LongWord(FProcessMask),false);
    FThreadMasks.Clear;
    for I := 0 to Pred(IDs.Count) do begin
      FThreadMasks.AddObject('$' + IntToHex(LongWord(IDs[I]),8),TObject(LongWord(FProcessMask)));
		  if FThreadAffinity then
    	  Self.SetThreadAffinityMask(LongWord(IDs[I]),LongWord(FProcessMask),false);
    end;
  finally
    CriticalSection.Leave;
    IDs.Free;
  end;
end;

procedure TProcessorAffinity.RemoveAllThreads;
var
  I: Integer;
begin
  FLastErrorMessage := '';
  CriticalSection.Enter;
  if FThreadAffinity then
	  for I := 0 to Pred(FThreadMasks.Count) do
  	  Self.SetThreadAffinityMask(LongWord(StrToInt(FThreadMasks.Strings[I])),LongWord(FProcessMask),false);
  FThreadMasks.Clear;
	CriticalSection.Leave;
end;

procedure TProcessorAffinity.SpreadThreads;
var
	I: Integer;
	Processor: TAffinityMaskEnum;
	Mask: TAffinityMaskSet;
begin
  FLastErrorMessage := '';
  if FThreadAffinity then begin
    Processor := 0;
    while not (Processor in FProcessMask) do
      Inc(Processor);
    CriticalSection.Enter;
    try
      for I := 0 to Pred(FThreadMasks.Count) do begin
        Mask := [Processor];
  	    FThreadMasks.Objects[I] := TObject(LongWord(Mask));
	      Self.SetThreadAffinityMask(LongWord(StrToInt(FThreadMasks.Strings[I])),LongWord(Mask),FExceptionOnError);
        Processor := (Processor + 1) mod (High(TAffinityMaskEnum) + 1);
        while not (Processor in FProcessMask) do
          Processor := (Processor + 1) mod (High(TAffinityMaskEnum) + 1);
      end
    finally
      CriticalSection.Leave;
    end;
  end;
end;

procedure TProcessorAffinity.Lock;
begin
  CriticalSection.Enter;
end;

procedure TProcessorAffinity.Unlock;
begin
  CriticalSection.Leave;
end;

procedure TProcessorAffinity.Loaded;
begin
  inherited Loaded;
  if FSetProcessMaskOnLoad and FProcessAffinity then
    SetProcessMask(FProcessMask);
end;

constructor TProcessorAffinity.Create(AOwner: TComponent);
var
  I: Integer;
  ProcessorN: Integer;
begin
  inherited Create(AOwner);
  DLLHandle := 0;
  OpenThread := nil;
  CriticalSection := TCriticalSection.Create;
  FThreadMasks := TStringList.Create;
  FThreadMasks.Sorted := True;
  FThreadMasks.Duplicates := dupError;
  PHandle := GetCurrentProcess;
  PID := GetCurrentProcessID;
  GetProcessAffinityMask(PHandle,LongWord(FProcessMask),LongWord(FSystemMask));
  ProcessorN := 0;
  for I := Low(TAffinityMaskEnum) to High(TAffinityMaskEnum) do
    if I in FSystemMask then
      Inc(ProcessorN);
  FProcessAffinity := (ProcessorN > 1) and (Win32Platform = VER_PLATFORM_WIN32_NT) and (Win32MajorVersion >= 4);
  FThreadAffinity := FProcessAffinity and (Win32Platform = VER_PLATFORM_WIN32_NT) and (Win32MajorVersion >= 5);
  if FThreadAffinity then begin
    DLLHandle := LoadLibrary('kernel32.dll');
    if DLLHandle <> 0 then
      @OpenThread := GetProcAddress(DLLHandle, 'OpenThread');
    if not Assigned(OpenThread) then
      FThreadAffinity := false;
  end;
end;

destructor TProcessorAffinity.Destroy;
begin
  RemoveAllThreads;
  if DLLHandle <> 0 then
    FreeLibrary(DLLHandle);
  FThreadMasks.Free;
  CriticalSection.Free;
  inherited Destroy;
end;


end.
